from ._tables import TABLE_ATOMNAME_ALIASES

# //////////////////////////////////////////////////////////////////////////////
class PDBPostProcess:
    # --------------------------------------------------------------------------
    @staticmethod
    def fix_prot_reslabels_aliases(pdb):
        ### [WIP] hardcoded for now for the prot.pdb example
        for res in pdb.residues:
            if res.resname == "HIS": res.resname = "HIE"


    # --------------------------------------------------------------------------
    @staticmethod
    def fix_prot_reslabels_termini(pdb):
        ### [WIP] needs improvement, e.g. for multiple chains
        residues = list(pdb.residues)
        nres = len(residues)
        for i,res in enumerate(residues):
            if i == 0:
                res.resname = f"N{res.resname}"
            elif i == nres - 1:
                res.resname = f"C{res.resname}"
                for atom in res.atoms:
                    if   atom.name == "O1": atom.name = "O"
                    elif atom.name == "O2": atom.name = "OXT"


    # --------------------------------------------------------------------------
    @staticmethod
    def fix_prot_atomlabels_aliases(pdb):
        ### [WIP] crudely hardcoded to at least support the prot.pdb example
        for atom in pdb.atoms:
            for resname, atom_aliases in TABLE_ATOMNAME_ALIASES.items():
                if atom.residue.resname.endswith(resname):
                    if atom.name in atom_aliases:
                        atom.name = atom_aliases[atom.name]
                    break


    # --------------------------------------------------------------------------
    @staticmethod
    def fix_rna_reslabels_termini(pdb):
        ### [WIP] improve this
        for res in pdb.residues:
            res_type = res.resname[0]
            if res_type not in "UCAG": continue

            isTerminus3 = False
            isTerminus5 = False
            for atom in res.atoms:
                if atom.name == "HO5'": isTerminus5 = True
                if atom.name == "HO3'": isTerminus3 = True

            if isTerminus3 and isTerminus5:
                res.resname = res_type + 'N'
            elif isTerminus3:
                res.resname = res_type + '3'
            elif isTerminus5:
                res.resname = res_type + '5'
            else:
                res.resname = res_type


# //////////////////////////////////////////////////////////////////////////////



import novemberff as nov

### [AS DEFINED IN] platforms/reference/include/SimTKOpenMMRealType.h
M_PI = 3.14159265358979323846
E_CHARGE = (1.602176634e-19)
AVOGADRO = (6.0221367e23)
EPSILON0 = (1e-6*8.8541878128e-12/(E_CHARGE*E_CHARGE*AVOGADRO)) # (e^2 Na/(kJ nm)) == (e^2/(kJ mol nm))


# //////////////////////////////////////////////////////////////////////////////
class FFNonBonded(nov.FFInteraction):
    ONE_4PI_EPS0 = 1/(4*M_PI*EPSILON0) # 138.935456

    # --------------------------------------------------------------------------
    def __init__(self,
        epsilon: float, sigma: float, charge: float | None,
        atom_type: "nov.FFAtomType"
    ):
        self.epsilon = epsilon
        self.sigma = sigma
        self.charge = charge
        self.type1 = atom_type


    # --------------------------------------------------------------------------
    def __repr__(self):
        return f"FFNonBonded(epsilon={self.epsilon}, sigma={self.sigma}, type={self.type1})"


    # --------------------------------------------------------------------------
    @classmethod
    def register_nonbonded(cls,
        epsilon: float, sigma: float, charge: float | None,
        str_type: str|None,
        str_class: str|None,
    ):
        strs_types   = (str_type,) # nov.FFInteraction expects tuples
        strs_classes = (str_class,)
        cls._decide_whether_using_names(strs_types, strs_classes)
        fftypes = cls._get_fftypes(strs_types, strs_classes)
        obj_nb = cls(epsilon, sigma, charge, *fftypes)
        cls._register_interaction(strs_types, strs_classes, obj_nb)


    # --------------------------------------------------------------------------
    @classmethod
    def get_nonbonded(cls, atom: nov.FFAtom) -> "nov.FFNonBonded":
        _current_map = cls._get_current_map()
        key = (
            atom.atom_type.name if cls._USING_NAMES else atom.atom_type.atom_class,
        ) # nov.FFInteraction expects tuples
        if key in _current_map:
            return _current_map[key]

        raise KeyError(f"{cls.__name__} type not found: {key}")


# //////////////////////////////////////////////////////////////////////////////



from pathlib import Path

# //////////////////////////////////////////////////////////////////////////////
class Utils:
    # --------------------------------------------------------------------------
    @staticmethod
    def combinations_proper_diheds(a0, a1, a2, a3):
        combinations = [
            (a0, a1, a2, a3),
            (a3, a2, a1, a0)
        ]
        masks = [
            (True,  True,  True,  True ),
            (True,  False, False, True ),
            (False, True,  True,  False),
        ]
        for mask in masks:
            for combo in combinations:
                yield combo, mask


    # --------------------------------------------------------------------------
    @staticmethod
    def combinations_improper_diheds(a0, a1, a2, a3):
        combinations = [
            (a0, a1, a2, a3),
            (a0, a1, a3, a2),
            (a2, a1, a0, a3),
            (a3, a1, a0, a2),
            (a2, a1, a3, a0),
            (a3, a1, a2, a0),
        ]
        masks = [
            (True,  True,  True,  True ),
            (True,  False, False, True ),
            (True,  False, True, True ),
        ]
        for mask in masks:
            for combo in combinations:
                yield combo, mask


    # --------------------------------------------------------------------------
    @staticmethod
    def solve_forcefield_path(ff_name: str) -> str:
        folder_data = Path(__file__).parent.parent / "_data"
        map_name2path = {
            "amber99sb": folder_data / "amber99sb.xml",
            "rna.ol3":   folder_data / "RNA.OL3.xml",
        }
        ff_path = map_name2path.get(ff_name.lower(), Path(ff_name))
        if not ff_path.exists():
            raise FileNotFoundError(f"Force field XML file not found: {ff_path}")
        return str(ff_path)


# //////////////////////////////////////////////////////////////////////////////



import novemberff as nov

# //////////////////////////////////////////////////////////////////////////////
class ForceField:
    def __init__(self, ff_data: dict):
        self._ffresidues     = ff_data["residues"]
        self.lj_14scale      = ff_data["LJ14SCALE"]
        self.coulomb_14scale = ff_data["COULOMB14SCALE"]
        self.one_4pi_eps0    = nov.FFNonBonded.ONE_4PI_EPS0

    # --------------------------------------------------------------------------
    @classmethod
    def from_xml(cls, path_xml):
        xml = nov.FFParserXML(path_xml)
        xml.parse()

        _safe_int   = lambda s: None if s is None else int(s)
        _safe_float = lambda s: None if s is None else float(s)

        node_ff        = xml.root.get_child_by_name("ForceField")
        node_types     = node_ff.get_child_by_name("AtomTypes")
        node_resids    = node_ff.get_child_by_name("Residues")
        node_bonds     = node_ff.get_child_by_name("HarmonicBondForce")
        node_angles    = node_ff.get_child_by_name("HarmonicAngleForce")
        node_diheds    = node_ff.get_child_by_name("PeriodicTorsionForce")
        node_nonbonded = node_ff.get_child_by_name("NonbondedForce")

        ff_data = {
            "types": {}, "residues": {},
            "LJ14SCALE":      float(node_nonbonded.get_attr("lj14scale")),
            "COULOMB14SCALE": float(node_nonbonded.get_attr("coulomb14scale")),
        }

        ffatomtypes = {}
        for child in node_types.children:
            name = child.get_attr("name")
            ffatomtypes[name] = nov.FFAtomType(
                atom_class = child.get_attr("class"),
                element    = child.get_attr("element"),
                mass       = float(child.get_attr("mass")),
                name       = name,
            )
        ff_data["types"] = ffatomtypes
        nov.FFBond.register_atomtypes(ffatomtypes)
        nov.FFAngle.register_atomtypes(ffatomtypes)
        nov.FFDihedral.register_atomtypes(ffatomtypes)
        nov.FFNonBonded.register_atomtypes(ffatomtypes)

        for residue_node in node_resids.children:
            resname = residue_node.get_attr("name")
            atom_nodes = filter(lambda c: c.tag_name == "Atom", residue_node.children)
            atoms = {
                node.get_attr("name") : nov.FFAtom(
                    name      = node.get_attr("name"),
                    charge    = _safe_float(node.get_attr("charge")),
                    atom_type = ffatomtypes[node.get_attr("type")],
                ) for node in atom_nodes
            }
            ff_data["residues"][resname] = nov.FFResidue(resname, atoms)

        for child in node_bonds.children:
            strs_types = (
                child.get_attr("type1"),
                child.get_attr("type2"),
            )
            strs_classes = (
                child.get_attr("class1"),
                child.get_attr("class2"),
            )
            nov.FFBond.register_bond(
                k      = float(child.get_attr("k")),
                length = float(child.get_attr("length")),
                strs_types = strs_types, strs_classes = strs_classes,
            )

        for child in node_angles.children:
            strs_types = (
                child.get_attr("type1"),
                child.get_attr("type2"),
                child.get_attr("type3"),
            )
            strs_classes = (
                child.get_attr("class1"),
                child.get_attr("class2"),
                child.get_attr("class3"),
            )
            nov.FFAngle.register_angle(
                k     = float(child.get_attr("k")),
                angle = float(child.get_attr("angle")),
                strs_types = strs_types, strs_classes = strs_classes,
            )

        for child in node_diheds.children:
            strs_types = (
                child.get_attr("type1"),
                child.get_attr("type2"),
                child.get_attr("type3"),
                child.get_attr("type4"),
            )
            strs_classes = (
                child.get_attr("class1"),
                child.get_attr("class2"),
                child.get_attr("class3"),
                child.get_attr("class4"),
            )
            nov.FFDihedral.register_dihed(
                isProper = child.tag_name == "Proper",
                ks = [
                    _safe_float(child.get_attr(f"k{i}")) for i in range(1,5)
                ],
                periodicities = [
                    _safe_int(child.get_attr(f"periodicity{i}")) for i in range(1,5)
                ],
                phases = [
                    _safe_float(child.get_attr(f"phase{i}")) for i in range(1,5)
                ],
                strs_types = strs_types, strs_classes = strs_classes,
            )

        for child in node_nonbonded.children:
            if child.tag_name != "Atom": continue
            str_type  = child.get_attr("type")
            str_class = child.get_attr("class")
            nov.FFNonBonded.register_nonbonded(
                epsilon  = float(child.get_attr("epsilon")),
                sigma    = float(child.get_attr("sigma")),
                charge   = _safe_float(child.get_attr("charge")),
                str_type = str_type, str_class = str_class,
            )

        return cls(ff_data)


    # --------------------------------------------------------------------------
    def omm2ff(self, atom) -> nov.FFAtom:
        """
        Convert an OpenMM atom to a FF atom.
        """
        ff_residue: nov.FFResidue = self._ffresidues[atom.residue.resname]
        return ff_residue.get_atom_by_name(atom.name)


    # --------------------------------------------------------------------------
    def get_ffbond(self, atom0, atom1) -> nov.FFBond:
        ff_a0 = self.omm2ff(atom0)
        ff_a1 = self.omm2ff(atom1)
        return nov.FFBond.get_bond(ff_a0, ff_a1)


    # --------------------------------------------------------------------------
    def get_ffangle(self, atom0, atom1, atom2) -> nov.FFAngle:
        ff_a0 = self.omm2ff(atom0)
        ff_a1 = self.omm2ff(atom1)
        ff_a2 = self.omm2ff(atom2)
        return nov.FFAngle.get_angle(ff_a0, ff_a1, ff_a2)


    # --------------------------------------------------------------------------
    def iter_ffpropers(self, atoms):
        for ordered_atoms,mask in nov.Utils.combinations_proper_diheds(*atoms):
            a0, a1, a2, a3 = ordered_atoms
            ff_a0 = self.omm2ff(a0)
            ff_a1 = self.omm2ff(a1)
            ff_a2 = self.omm2ff(a2)
            ff_a3 = self.omm2ff(a3)
            dihed = nov.FFDihedral.get_dihed(ff_a0, ff_a1, ff_a2, ff_a3, mask)
            if dihed is not None:
                ordered_idxs = (a0.index, a1.index, a2.index, a3.index)
                yield dihed, ordered_idxs


    # --------------------------------------------------------------------------
    def iter_ffimpropers(self, atoms):
        ### improper dihedrals in amber should have "a2" at the center (in theory)
        ### https://ambermd.org/doc12/Amber24.pdf#page=279
        ### but in practice, for openmm they have "a1" at the center
        ### graph_bonds.py reports "a0" in the center, so switch a0 and a1

        for ordered_atoms,mask in nov.Utils.combinations_improper_diheds(*atoms):
            a0, a1, a2, a3 = ordered_atoms
            ff_a0 = self.omm2ff(a1) # swap between 0 and 1 is intentional (see explanation above)
            ff_a1 = self.omm2ff(a0)
            ff_a2 = self.omm2ff(a2)
            ff_a3 = self.omm2ff(a3)
            dihed = nov.FFDihedral.get_dihed(ff_a0, ff_a1, ff_a2, ff_a3, mask)
            if dihed is not None:
                ordered_idxs = (a0.index, a1.index, a2.index, a3.index)
                yield dihed, ordered_idxs


    # --------------------------------------------------------------------------
    def get_ffnonbonded(self, a0, a1):
            ff_a0 = self.omm2ff(a0)
            ff_a1 = self.omm2ff(a1)
            ff_nb0 = nov.FFNonBonded.get_nonbonded(ff_a0)
            ff_nb1 = nov.FFNonBonded.get_nonbonded(ff_a1)

            ### charges are sometimes stored in the Residue.Atom XML-nodes (e.g. in RNA.OL3.xml)
            ### other times, they are stored in the Nonbonded XML-nodes (e.g. in amber99sb.xml)
            charge0 = ff_a0.charge if ff_nb0.charge is None else ff_nb0.charge
            charge1 = ff_a1.charge if ff_nb1.charge is None else ff_nb1.charge
            return ff_nb0, ff_nb1, charge0, charge1


# //////////////////////////////////////////////////////////////////////////////



from enum import Enum, auto

import novemberff as nov

# ////////////////////////////////////////////////////////////////////////////////
class State(Enum):
    START = auto()
    TAG_NAME = auto()
    TAG_CONTENT = auto()
    ATTRIBUTE_KEY = auto()
    ATTRIBUTE_VAL = auto()
    END = auto()


# ////////////////////////////////////////////////////////////////////////////////
class XMLNode:
    def __init__(self, tag_name, attributes, tag_content):
        self.tag_name: str = tag_name
        self.attributes: dict[str, str] = attributes
        self.tag_content: str = tag_content
        self.children: list["XMLNode"] = []

    # --------------------------------------------------------------------------
    def __repr__(self):
        return f"XMLNode(name='{self.tag_name}', nattrs={len(self.attributes)}, content='{self.tag_content}', nchildren={len(self.children)})"

    # --------------------------------------------------------------------------
    def add_child(self, node):
        self.children.append(node)

    # --------------------------------------------------------------------------
    def get_child_by_name(self, name: str) -> "XMLNode | None":
        for child in self.children:
            if child.tag_name == name:
                return child

    # --------------------------------------------------------------------------
    def has_attr(self, key: str) -> bool:
        return key in self.attributes.keys()

    # --------------------------------------------------------------------------
    def get_attr(self, key: str) -> str | None:
        if self.has_attr(key):
            return self.attributes[key]


# ////////////////////////////////////////////////////////////////////////////////
class FFParserXML(nov.ForceFieldParser):
    def __init__(self, path_xml):
        self.root = XMLNode('', {}, '')
        self._path_xml = path_xml
        with open(path_xml, 'r') as file:
            self._xml_content = file.read()


    # --------------------------------------------------------------------------
    def parse(self):
        node = self.root
        parent_chain = []
        for line in self._xml_content.splitlines():
            tag_name, attributes, tag_content, ended = self.parse_xml_line(line)

            ### CASE 0: end-tag only
            if not tag_name:
                node = parent_chain.pop()
                continue

            child = XMLNode(tag_name, attributes, tag_content)

            ### CASE 1: empty-element-tag or (start-tag + content + end-tag) in a single line
            node.add_child(child)
            if ended: continue

            ### CASE 2: start-tag only
            parent_chain.append(node)
            node = child


    # --------------------------------------------------------------------------
    @classmethod
    def parse_xml_line(cls, line: str):
        state = State.START
        tag_name = ""
        tag_content = ""
        attributes = {}
        current_key = ""

        line = line.strip()
        for i,c in enumerate(line):
            if state == State.START:
                if line[i+1] == "/":
                    state = State.END
                elif c == "<":
                    state = State.TAG_NAME
                continue

            if state == State.TAG_NAME:
                if c == ">":
                    state = State.TAG_CONTENT
                elif c == " ":
                    state = State.ATTRIBUTE_KEY
                else:
                    tag_name += c
                continue

            if state == State.TAG_CONTENT:
                if c == "<":
                    state = State.END
                else:
                    tag_content += c
                continue

            if state == State.ATTRIBUTE_KEY:
                if c == "=":
                    attributes[current_key] = ""
                elif c == "\"":
                    state = State.ATTRIBUTE_VAL
                elif c == ">":
                    state = State.TAG_CONTENT
                elif c == "/":
                    state = State.END
                elif c != " ":
                    current_key += c
                continue

            if state == State.ATTRIBUTE_VAL:
                if c == "\"":
                    current_key = ""
                    state = State.ATTRIBUTE_KEY
                elif c != " ":
                    attributes[current_key] += c
                continue

        return tag_name, attributes, tag_content, state == State.END


# ////////////////////////////////////////////////////////////////////////////////



import numpy as np
from pathlib import Path
import MDAnalysis as mda

import novemberff as nov

# //////////////////////////////////////////////////////////////////////////////
class EnergyCalculator:
    def __init__(self, path_pdb: Path, forcefield: str | Path):
        self._path_pdb = path_pdb

        self._pdb = mda.Universe(str(path_pdb), to_guess = ["elements", "bonds"])
        self._bgraph = nov.BondGraph(
            self._pdb.atoms, self._pdb.bonds,
            coords = self._pdb.atoms.positions
        )
        path_xml = nov.Utils.solve_forcefield_path(forcefield)
        self._forcefield = nov.ForceField.from_xml(path_xml)

        self._natoms     = len(self._bgraph.atoms)
        self._nbonds     = len(self._bgraph.bonds)
        self._nangles    = len(self._bgraph.angles)
        self._npropers   = len(self._bgraph.proper_diheds)
        self._nimpropers = len(self._bgraph.improper_diheds)
        self._nnonbonded = len(self._bgraph.nonbonded_pairs)

        self._arr_bond_energies     = np.zeros(self._nbonds)
        self._arr_angle_energies    = np.zeros(self._nangles)
        self._arr_proper_energies   = np.zeros(self._npropers)
        self._arr_improper_energies = np.zeros(self._nimpropers)
        self._arr_lennardj_energies = np.zeros(self._nnonbonded)
        self._arr_coulomb_energies  = np.zeros(self._nnonbonded)


    # --------------------------------------------------------------------------
    @classmethod
    def with_prot_ff(cls, path_pdb: Path) -> "EnergyCalculator":
        """Initialize EnergyCalculator with a default protein force field (Amber99SB)."""
        return cls(path_pdb, "amber99sb")


    # --------------------------------------------------------------------------
    @classmethod
    def with_rna_ff(cls, path_pdb: Path) -> "EnergyCalculator":
        """Initialize EnergyCalculator with a default RNA force field (RNA.OL3)."""
        return cls(path_pdb, "rna.ol3")


    # --------------------------------------------------------------------------
    def calc_energies_prot(self) -> "EnergyCalculator":
        nov.PDBPostProcess.fix_prot_atomlabels_aliases(self._pdb)
        nov.PDBPostProcess.fix_prot_reslabels_aliases(self._pdb)
        nov.PDBPostProcess.fix_prot_reslabels_termini(self._pdb)
        self._calc_energies()
        return self


    # --------------------------------------------------------------------------
    def calc_energies_rna(self) -> "EnergyCalculator":
        nov.PDBPostProcess.fix_rna_reslabels_termini(self._pdb)
        self._calc_energies()
        return self


    # --------------------------------------------------------------------------
    def get_array_ebond(self):     return self._arr_bond_energies
    def get_array_eangle(self):    return self._arr_angle_energies
    def get_array_eproper(self):   return self._arr_proper_energies
    def get_array_eimproper(self): return self._arr_improper_energies
    def get_array_elennardj(self): return self._arr_lennardj_energies
    def get_array_ecoulomb(self):  return self._arr_coulomb_energies

    def get_array_edihed(self):
        return np.concat((self.get_array_eproper(), self.get_array_eimproper()))

    def get_array_enonbonded(self):
        return np.concat((self.get_array_elennardj(), self.get_array_ecoulomb()))


    # --------------------------------------------------------------------------
    def get_sum_ebond(self):     return np.sum(self._arr_bond_energies)
    def get_sum_eangle(self):    return np.sum(self._arr_angle_energies)
    def get_sum_eproper(self):   return np.sum(self._arr_proper_energies)
    def get_sum_eimproper(self): return np.sum(self._arr_improper_energies)
    def get_sum_elennardj(self): return np.sum(self._arr_lennardj_energies)
    def get_sum_ecoulomb(self):  return np.sum(self._arr_coulomb_energies)

    def get_sum_edihed(self):
        return self.get_sum_eproper() + self.get_sum_eimproper()

    def get_sum_enonbonded(self):
        return self.get_sum_elennardj() + self.get_sum_ecoulomb()


    # --------------------------------------------------------------------------
    def display_energies(self):
        print(f">>> Amber energies (NovemberFF) for '{self._path_pdb}':")
        print("  BondEnergy:",  self.get_sum_ebond())
        print("  AngleEnergy:", self.get_sum_eangle())
        print("  DihedEnergy:", self.get_sum_edihed())
        print("  Nonbonded:",   self.get_sum_enonbonded())
        print("  ... LennardJones:", self.get_sum_elennardj())
        print("  ... Coulomb:",      self.get_sum_ecoulomb())


    # --------------------------------------------------------------------------
    def save_energy_arrays(self, folder_output: str | Path):
        folder_output = Path(folder_output)
        folder_output.mkdir(parents = True, exist_ok = True)
        np.save(folder_output / f"bonds.npy",     self._arr_bond_energies    )
        np.save(folder_output / f"angles.npy",    self._arr_angle_energies   )
        np.save(folder_output / f"propers.npy",   self._arr_proper_energies  )
        np.save(folder_output / f"impropers.npy", self._arr_improper_energies)
        np.save(folder_output / f"lennardjs.npy", self._arr_lennardj_energies)
        np.save(folder_output / f"coulombs.npy",  self._arr_coulomb_energies )


    # --------------------------------------------------------------------------
    def _calc_energies(self):
        ##### novemberff_e = (1 ± ε) * openmm_e
        self._calc_ebonded()   # ε = 1e-14
        self._calc_eangles()   # ε = 1e-15
        self._calc_ediheds()   # ε = 1e-4
        self._calc_nonbonded() # ε = 1e-6


    # --------------------------------------------------------------------------
    def _calc_ebonded(self):
        ###### [ORIGINAL SOURCE] platforms/reference/src/SimTKReference/ReferenceHarmonicBondlxn.cpp@111
        for i,(a0,a1) in enumerate(self._bgraph.bonds):
            if (a0.element == 'H') or (a1.element == 'H'): # apply HBond constraints
                continue

            ff_bond  = self._forcefield.get_ffbond(a0, a1)
            distance = self._bgraph.calc_dist_2atoms(a0.index, a1.index)
            energy = 0.5 * ff_bond.k * (ff_bond.length - distance)**2
            self._arr_bond_energies[i] = energy


    # --------------------------------------------------------------------------
    def _calc_eangles(self):
        for i,(a0,a1,a2) in enumerate(self._bgraph.angles):
            ff_angle = self._forcefield.get_ffangle(a0, a1, a2)
            angle  = self._bgraph.calc_angle_3atoms(a0.index, a1.index, a2.index)
            energy = 0.5 * ff_angle.k * (angle - ff_angle.angle)**2
            self._arr_angle_energies[i] = energy


    # --------------------------------------------------------------------------
    def _calc_ediheds(self):
        def _edihed_contributions(ff_dihed: nov.FFDihedral, ordered_atoms, is_proper):
            a0,a1,a2,a3 = ordered_atoms
            angle = self._bgraph.calc_dihed_4atoms(a0, a1, a2, a3)
            e1 = ff_dihed.calc_energy(angle, contributor = 1, proper = is_proper)
            e2 = ff_dihed.calc_energy(angle, contributor = 2, proper = is_proper)
            e3 = ff_dihed.calc_energy(angle, contributor = 3, proper = is_proper)
            e4 = ff_dihed.calc_energy(angle, contributor = 4, proper = is_proper)
            return e1 + e2 + e3 + e4

        for i,atoms in enumerate(self._bgraph.proper_diheds):
            for ff_dihed, ordered_atoms in self._forcefield.iter_ffpropers(atoms):
                energy = _edihed_contributions(ff_dihed, ordered_atoms, is_proper = True)
                if energy:
                    self._arr_proper_energies[i] = energy
                    break

        for i,atoms in enumerate(self._bgraph.improper_diheds):
            for ff_dihed, ordered_atoms in self._forcefield.iter_ffimpropers(atoms):
                energy = _edihed_contributions(ff_dihed, ordered_atoms, is_proper = False)
                if energy:
                    self._arr_improper_energies[i] = energy
                    break


    # --------------------------------------------------------------------------
    def _calc_nonbonded(self):
        ### [ORIGINAL SOURCE] platforms/cpu/src/CpuKernels.cpp@CpuCalcNonbondedForceKernel::computeParameters
        ### [ORIGINAL SOURCE] platforms/cpu/include/CpuNonbondedForceFvec.h@CpuNonbondedForceFvec::calculateBlockIxnImpl
        ### [ORIGINAL SOURCE] openmmapi/src/NonbondedForce.cpp@NonbondedForce::createExceptionsFromBonds
        ### [ORIGINAL SOURCE] platforms/reference/src/SimTKReference/ReferenceLJCoulomb14.cpp@ReferenceLJCoulomb14::calculateBondIxn
        for i,(a0,a1,bond_edge_dist) in enumerate(self._bgraph.nonbonded_pairs):
            ff_nb0, ff_nb1, charge0, charge1 = self._forcefield.get_ffnonbonded(a0, a1)

            inverse_r = 1 / self._bgraph.calc_dist_2atoms(a0.index, a1.index)
            sigma = 0.5 * (ff_nb0.sigma + ff_nb1.sigma)
            epsilon = 4.0 * np.sqrt(ff_nb0.epsilon * ff_nb1.epsilon)
            charge = charge0 * charge1
            sig6 = (inverse_r * sigma)**6

            energy_lj = epsilon * sig6 * (sig6 - 1.0)
            energy_coul = inverse_r * charge * self._forcefield.one_4pi_eps0

            if bond_edge_dist == nov.BondEdgeDist.mid: # apply 1-4 scaling to atoms separated by 3 bonds
                energy_lj   *= self._forcefield.lj_14scale
                energy_coul *= self._forcefield.coulomb_14scale # AKA "epsilon_factor"

            self._arr_lennardj_energies[i] = energy_lj
            self._arr_coulomb_energies[i] = energy_coul


# //////////////////////////////////////////////////////////////////////////////



from ._misc.utils import Utils

from ._parsers.ffparser import ForceFieldParser
from ._parsers.ffparser_xml import FFParserXML, XMLNode
from ._parsers.pdb_postprocess import PDBPostProcess

from ._forcefield.particles import FFAtomType, FFAtom, FFResidue
from ._forcefield.interaction import FFInteraction
from ._forcefield.bonds import FFBond
from ._forcefield.angles import FFAngle
from ._forcefield.dihedrals import FFDihedral
from ._forcefield.nonbonded import FFNonBonded
from ._forcefield.forcefield import ForceField

from ._core.bond_graph import BondGraph, BondEdgeDist
from ._core.energy_calculator import EnergyCalculator



import math

import novemberff as nov

# //////////////////////////////////////////////////////////////////////////////
class FFDihedral(nov.FFInteraction):
    def __init__(self,
        isProper: bool,
        ks: tuple[float, float, float, float],
        periodicities: tuple[int, int, int, int],
        phases: tuple[float, float, float, float],
        type1: "nov.FFAtomType", type2: "nov.FFAtomType",
        type3: "nov.FFAtomType", type4: "nov.FFAtomType",
    ):
        self.isProper = isProper
        self.k1 = ks[0]
        self.k2 = ks[1]
        self.k3 = ks[2]
        self.k4 = ks[3]
        self.periodicity1 = periodicities[0]
        self.periodicity2 = periodicities[1]
        self.periodicity3 = periodicities[2]
        self.periodicity4 = periodicities[3]
        self.phase1 = phases[0]
        self.phase2 = phases[1]
        self.phase3 = phases[2]
        self.phase4 = phases[3]
        self.type1 = type1
        self.type2 = type2
        self.type3 = type3
        self.type4 = type4


    # --------------------------------------------------------------------------
    def __repr__(self):
        return f"FFDihedral({self.type1}, {self.type2}, {self.type3}, {self.type4})"


    # --------------------------------------------------------------------------
    @classmethod
    def register_dihed(cls,
        isProper: bool,
        ks: tuple[float, float, float, float],
        periodicities: tuple[int, int, int, int],
        phases: tuple[float, float, float, float],
        strs_types:   tuple[str|None, str|None, str|None],
        strs_classes: tuple[str|None, str|None, str|None],
    ):
        cls._decide_whether_using_names(strs_types, strs_classes)
        fftypes = cls._get_fftypes(strs_types, strs_classes)
        obj_dihed = cls(isProper, ks, periodicities, phases, *fftypes)
        cls._register_interaction(strs_types, strs_classes, obj_dihed)


    # --------------------------------------------------------------------------
    def calc_energy(self, angle: float, contributor: int, proper: bool) -> float:
        match contributor:
            case 1:
                k = self.k1
                phase = self.phase1
                periodicity = self.periodicity1
            case 2:
                k = self.k2
                phase = self.phase2
                periodicity = self.periodicity2
            case 3:
                k = self.k3
                phase = self.phase3
                periodicity = self.periodicity3
            case 4:
                k = self.k4
                phase = self.phase4
                periodicity = self.periodicity4

        if k is None: return 0 # skip irrelevant contributors
        if self.isProper ^ proper: return 0

        ### [ORIGINAL SOURCE] platforms/reference/src/SimTKReference/ReferenceProperDihedralBond.cpp@ReferenceProperDihedralBond::calculateBondIxn
        delta_angle = periodicity * angle - phase
        return k * (1.0 + math.cos(delta_angle))


    # --------------------------------------------------------------------------
    @classmethod
    def get_dihed(cls,
        atom1: nov.FFAtom, atom2: nov.FFAtom,
        atom3: nov.FFAtom, atom4: nov.FFAtom,
        mask: tuple[bool, bool, bool, bool]
    ) -> "nov.FFDihedral | None":
        def _id(a: nov.FFAtom) -> str:
            return a.atom_type.name if cls._USING_NAMES else a.atom_type.atom_class

        key = (
            _id(atom1) if mask[0] else '',
            _id(atom2) if mask[1] else '',
            _id(atom3) if mask[2] else '',
            _id(atom4) if mask[3] else '',
        )
        return cls._get_current_map().get(key)


# //////////////////////////////////////////////////////////////////////////////



from abc import ABC

import novemberff as nov

# //////////////////////////////////////////////////////////////////////////////
class FFInteraction(ABC):
    _ATOMTYPES_BY_NAME  = {}
    _ATOMTYPES_BY_CLASS = {}
    _USING_NAMES = None # will be set to True/False upon first interaction registered
    _MAP_INTERACTION_BY_NAME  = {}
    _MAP_INTERACTION_BY_CLASS = {}

    # --------------------------------------------------------------------------
    @classmethod
    def register_atomtypes(cls, atomtypes: dict[str, nov.FFAtomType]):
        cls._ATOMTYPES_BY_NAME = atomtypes
        cls._ATOMTYPES_BY_CLASS = {a.atom_class: a for a in atomtypes.values()}


    # --------------------------------------------------------------------------
    @classmethod
    def _decide_whether_using_names(cls,
        strs_types: tuple[str|None,...], strs_classes: tuple[str|None,...]
    ):
        if cls._USING_NAMES is not None: return

        if not((strs_types[0] is None)^(strs_classes[0] is None)):
            raise ValueError("Forcefield must consistently provide either atom types or atom classes for its interactions.")

        cls._USING_NAMES = strs_classes[0] is None


    # --------------------------------------------------------------------------
    @classmethod
    def _register_interaction(cls,
        strs_types: tuple[str|None,...], strs_classes: tuple[str|None,...],
        obj_interaction: "FFInteraction"
    ):
        if cls._USING_NAMES:
            cls._MAP_INTERACTION_BY_NAME[strs_types] = obj_interaction
            return
        cls._MAP_INTERACTION_BY_CLASS[strs_classes] = obj_interaction


    # --------------------------------------------------------------------------
    @classmethod
    def _get_current_map(cls):
        return cls._MAP_INTERACTION_BY_NAME if cls._USING_NAMES else cls._MAP_INTERACTION_BY_CLASS


    # --------------------------------------------------------------------------
    @classmethod
    def _get_fftypes(cls,
        str_types: tuple[str,...], str_classes: tuple[str,...]
    ) -> list["nov.FFAtomType"]:
        return [
            cls._ATOMTYPES_BY_NAME.get(t) for t in str_types
        ] if cls._USING_NAMES else [
            cls._ATOMTYPES_BY_CLASS.get(c) for c in str_classes
        ]


# //////////////////////////////////////////////////////////////////////////////



import math
import numpy as np
from enum import Enum, auto

import novemberff as nov

# //////////////////////////////////////////////////////////////////////////////
class BondEdgeDist(Enum):
    short = auto()
    mid   = auto()
    long  = auto()


# //////////////////////////////////////////////////////////////////////////////
class BondGraph:
    LARGE_INT = 9

    # --------------------------------------------------------------------------
    def __init__(self, atoms, bonds, coords):
        self.atoms = list(atoms)
        self.coords = list(coords)
        self._bond_matrix = np.full((len(self.atoms), len(self.atoms)), self.LARGE_INT, dtype = int)

        self.bonds = [] # [(a1, a2), ...]
        self.angles = [] # [(a1, a2, a3), ...]
        self.proper_diheds = [] # [(a1, a2, a3, a4), ...]
        self.improper_diheds = set() # {(a1, a2, a3, a4), ...}
        self.nonbonded_pairs: list[tuple] # [(a1, a2, BondEdgeDist), ...]


        # ----------------------------------------------------------------------
        for bond in bonds: # init with direct bonds
            a1,a2 = bond
            self._bond_matrix[a1.index, a2.index] = 1
            self._bond_matrix[a2.index, a1.index] = 1

        # ----------------------------------------------------------------------
        for i,row_i in enumerate(self._bond_matrix):
            self._bond_matrix[i][i] = 0
            for j,val_j in enumerate(row_i[i:], start = i):
                if val_j == 1:
                    self.bonds.append((self.atoms[i], self.atoms[j]))

        # ----------------------------------------------------------------------
        for idx_bond,bond0 in enumerate(self.bonds[:-1]):
            ### bonds are of the form (i-j)
            a0i, a0j = bond0 # (ij)
            for bond1 in self.bonds[idx_bond+1:]:
                a1k, a1l = bond1 # (kl)

                ### (i*) and (k*)
                if   a0j == a1l: self._add_angle(a0i, a0j, a1k)

                ### (i*) and (*l)
                elif a0j == a1k: self._add_angle(a0i, a0j, a1l)

                ### (*j) and (k*)
                elif a0i == a1l: self._add_angle(a0j, a0i, a1k)

                ### (*j) and (*l)
                elif a0i == a1k: self._add_angle(a0j, a0i, a1l)

        # ----------------------------------------------------------------------
        for idx_angle,angle0 in enumerate(self.angles[:-1]):
            ### angles are of the form (i-j-k)
            a0i, a0j, a0k = angle0 # (ijk)
            angle0_r = a0k, a0j, a0i # (kji)
            for angle1 in self.angles[idx_angle+1:]:
                a1l, a1m, a1n = angle1 # (lmn)

                ### (*jk) and (l*n)
                if   a0i == a1m:
                    ### (**k) and (**n)
                    if   a0j == a1l: self._add_proper_dihed(*angle0_r, a1n)

                    ### (**k) and (l**)
                    elif a0j == a1n: self._add_proper_dihed(*angle0_r, a1l)

                ### (ij*) and (l*n)
                elif a0k == a1m:
                    ### (i**) and (**n)
                    if   a0j == a1l: self._add_proper_dihed(*angle0, a1n)

                    ### (i**) and (l**)
                    elif a0j == a1n: self._add_proper_dihed(*angle0, a1l)

                ### (i*k) and (l*n)
                elif a0j == a1m:
                    ### (i**) and (**n)
                    if   a0k == a1l: self._add_improper_dihed(*angle0, a1n)

                    ### (i**) and (l**)
                    elif a0k == a1n: self._add_improper_dihed(*angle0, a1l)

                    ### (**k) and (**n)
                    elif a0i == a1l: self._add_improper_dihed(*angle0_r, a1n)

                    ### (**k) and (l**)
                    elif a0i == a1n: self._add_improper_dihed(*angle0_r, a1l)

                ### (*jk) and (*mn)
                elif a0i == a1l: self._set_cell_value(a0k, a1n, 4)

                ### (ij*) and (*mn)
                elif a0k == a1l: self._set_cell_value(a0i, a1n, 4)

                ### (*jk) and (lm*)
                elif a0i == a1n: self._set_cell_value(a0k, a1l, 4)

                ### (ij*) and (lm*)
                elif a0k == a1n: self._set_cell_value(a0i, a1l, 4)

        # ----------------------------------------------------------------------
        self.nonbonded_pairs = list(filter(
            lambda t: t[2] != BondEdgeDist.short, # skip atoms separated by less than 3 bonds
            (
                (a0, a1, self._get_bond_edge_dist(a0.index, a1.index))
                for idx,a0 in enumerate(self.atoms[:-1])
                for a1 in self.atoms[idx+1:]
            )
        ))


    # ----------------------------------------------------------------------
    def calc_dist_2atoms(self, idx0, idx1):
        dist = self.coords[idx0] - self.coords[idx1]
        dist /= 10 # angstroms to nanometers
        return np.linalg.norm(dist)


    # ----------------------------------------------------------------------
    def calc_angle_3atoms(self, idx0, idx1, idx2):
        vec0 = self.coords[idx0] - self.coords[idx1]
        vec1 = self.coords[idx2] - self.coords[idx1]
        cosine = np.dot(vec0, vec1) / (math.sqrt(np.dot(vec0, vec0) * np.dot(vec1, vec1)))
        return math.acos(cosine) # radians


    # ----------------------------------------------------------------------
    def calc_dihed_4atoms(self, idx0, idx1, idx2, idx3):
        # Assumes the following order of atoms:
        # PROPER DIHEDRALS
        #   1   3 :   1 1   3
        #  / \ /  :  / \ \ /
        # 0   2   : 0   2 2
        #
        # IMPROPER DIHEDRALS
        #   2   :   2 2   [WIP] do the operations still make sense geometrically?
        #   |   :   | |
        #   0   :   0 0
        #  / \  :  /   \
        # 1   3 : 1     3

        vec0 = self.coords[idx3] - self.coords[idx2]
        vec1 = self.coords[idx1] - self.coords[idx2]
        vec2 = self.coords[idx1] - self.coords[idx0]
        n0 = np.cross(vec0, vec1)
        n1 = np.cross(vec1, vec2)
        cosine = np.dot(n0, n1) / (math.sqrt(np.dot(n0, n0) * np.dot(n1, n1)))

        sign = 1 if np.dot(vec0, n1) > 0 else -1
        cosine = max(-1.0, min(1.0, cosine))
        return sign * math.acos(cosine)


    # ----------------------------------------------------------------------
    def _add_angle(self, a0, a1, a2):
        self.angles.append((a0, a1, a2))
        self._set_cell_value(a0, a2, 2)
        self._set_cell_value(a2, a0, 2)


    # ----------------------------------------------------------------------
    def _add_proper_dihed(self, a0, a1, a2, a3):
        self.proper_diheds.append((a0, a1, a2, a3))
        self._set_cell_value(a0, a3, 3)
        self._set_cell_value(a3, a0, 3)


    # ----------------------------------------------------------------------
    def _add_improper_dihed(self, a0, a1, a2, a3):
        #   2
        #   |
        #   1
        #  / \
        # 0   3
        a0, a2, a3 = sorted((a0, a2, a3), key = lambda atom: atom.name)
        self.improper_diheds.add((a0, a1, a2, a3))


    # ----------------------------------------------------------------------
    def _set_cell_value(self, a0, a1, value):
        i = a0.index
        j = a1.index
        self._bond_matrix[i][j] = min(value, self._bond_matrix[i][j])
        self._bond_matrix[j][i] = min(value, self._bond_matrix[j][i])


    # ----------------------------------------------------------------------
    def _get_bond_edge_dist(self, i0, i1):
        dist = self._bond_matrix[i0][i1]
        if dist >  3: return BondEdgeDist.long
        if dist == 3: return BondEdgeDist.mid
        return BondEdgeDist.short


################################################################################
if __name__ == "__main__":
    class TestAtom:
        def __init__(self, index):
            self.index = index

    class TestBond:
        def __init__(self, atom1, atom2):
            self.atom1 = atom1
            self.atom2 = atom2

    atoms = [TestAtom(i) for i in range(10)]
    bonds = [
        TestBond(atoms[0], atoms[1]),
        TestBond(atoms[0], atoms[4]),
        TestBond(atoms[1], atoms[7]),
        TestBond(atoms[2], atoms[3]),
        TestBond(atoms[2], atoms[4]),
        TestBond(atoms[2], atoms[7]),
        TestBond(atoms[4], atoms[5]),
        TestBond(atoms[5], atoms[6]),
        TestBond(atoms[6], atoms[8]),
        TestBond(atoms[8], atoms[9]),
    ]
    coords = np.zeros((8, 3))

    bond_graph = BondGraph(atoms, bonds, coords)
    for a1,a2 in bond_graph.bonds:
        print(a1.index, a2.index)

    for a1,a2,a3 in bond_graph.angles:
        print(a1.index, a2.index, a3.index)

    for a0,a1,a2,a3 in bond_graph.proper_diheds:
        print(a0.index, a1.index, a2.index, a3.index)

    print(bond_graph._bond_matrix)


################################################################################



from setuptools import setup, find_packages

setup(
    name="novemberff",
    version="0.3.0",
    description="Custom implementation of the AMBER forcefield, focused on simplicity and verbose outputs.",
    keywords="Amber Forcefield MolecularDynamics MD Simulation",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="DiegoBarMor",
    author_email="diegobarmor42@gmail.com",
    url="https://github.com/diegobarmor/novemberff",
    license="MIT",
    packages=find_packages(),
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)



import novemberff as nov

# //////////////////////////////////////////////////////////////////////////////
class FFAngle(nov.FFInteraction):
    def __init__(self,
        k: float, angle: float,
        type1: "nov.FFAtomType", type2: "nov.FFAtomType",
        type3: "nov.FFAtomType"
    ):
        self.k = k
        self.angle = angle
        self.type1 = type1
        self.type2 = type2
        self.type3 = type3


    # --------------------------------------------------------------------------
    def __repr__(self):
        return f"FFAngle(k={self.k}, angle={self.angle}, type1={self.type1}, type2={self.type2}, type3={self.type3})"


    # --------------------------------------------------------------------------
    @classmethod
    def register_angle(cls,
        k: float, angle: float,
        strs_types:   tuple[str|None, str|None, str|None],
        strs_classes: tuple[str|None, str|None, str|None],
    ):
        cls._decide_whether_using_names(strs_types, strs_classes)
        fftypes = cls._get_fftypes(strs_types, strs_classes)
        obj_angle = cls(k, angle, *fftypes)
        cls._register_interaction(strs_types, strs_classes, obj_angle)


    # --------------------------------------------------------------------------
    @classmethod
    def get_angle(cls, atom1: nov.FFAtom, atom2: nov.FFAtom, atom3: nov.FFAtom) -> "nov.FFAngle":
        _current_map = cls._get_current_map()
        for key in cls._iter_possible_keys(atom1, atom2, atom3):
            if key in _current_map:
                return _current_map[key]

        raise KeyError(f"{cls.__name__} type not found: {key}")


    # --------------------------------------------------------------------------
    @classmethod
    def _iter_possible_keys(cls, atom1: nov.FFAtom, atom2: nov.FFAtom, atom3: nov.FFAtom):
        """Return possible key tuples for looking up angle parameters."""
        atype1 = atom1.atom_type
        atype2 = atom2.atom_type
        atype3 = atom3.atom_type
        if cls._USING_NAMES:
            yield (atype1.name, atype2.name, atype3.name)
            yield (atype3.name, atype2.name, atype1.name)
        else:
            yield (atype1.atom_class, atype2.atom_class, atype3.atom_class)
            yield (atype3.atom_class, atype2.atom_class, atype1.atom_class)


# //////////////////////////////////////////////////////////////////////////////
